@@ -60,15 +60,51 @@ class IssueManagerTests(TestCase):
 
         self.assertTrue(Issue.objects.is_wont_fix(name=mi.name))
 
-    def test_resolve_issue(self):
-        name1 = 'a-name'
-        name2 = 'another-name'
-        i = G(Issue, name=name1, status=IssueStatus.Resolved.value)
-        i2 = G(Issue, name=name2, status=IssueStatus.Resolved.value)
+    def test_maybe_open_issue_when_none_exists(self):
+        """
+        Verify that maybe_open_issue will create a new Issue when none like it exists.
+        """
+        (issue, created) = Issue.objects.maybe_open_issue(name='falafel')
+        self.assertTrue(created)
+        self.assertEqual(IssueStatus.Open.value, Issue.objects.get(name=issue.name).status)
 
-        Issue.objects.filter(name=name1).resolve()
-        self.assertEqual(IssueStatus.Resolved.value, Issue.objects.get(pk=i.pk).status)
-        self.assertEqual(IssueStatus.Open.value, Issue.objects.get(pk=i2.pk).status)
+    def test_maybe_open_issue_when_it_is_marked_as_wont_fix(self):
+        """
+        Verify that maybe_open_issue will not create or return an Issue when it exists and
+        is marked as WONT_FIX.
+        """
+        issue = G(Issue, status=IssueStatus.Wont_fix.value)
+        self.assertEqual((None, False), Issue.objects.maybe_open_issue(name=issue.name))
+        self.assertEqual(IssueStatus.Wont_fix.value, Issue.objects.get(pk=issue.pk).status)
+        self.assertEqual(1, Issue.objects.filter(name=issue.name).count())
+
+    def test_maybe_open_issue_returns_already_open_issue(self):
+        """
+        Verify that maybe_open_issue will return a the extant Issue of hte provided name
+        when it is open.
+        """
+        issue = G(Issue, status=IssueStatus.Open.value)
+        (issue2, created) = Issue.objects.maybe_open_issue(name=issue.name)
+        self.assertFalse(created)
+        self.assertEqual(IssueStatus.Open.value, Issue.objects.get(pk=issue.pk).status)
+        self.assertEqual(1, Issue.objects.filter(name=issue.name).count())
+
+    def maybe_open_issue_when_it_is_marked_as_resolved(self):
+        """
+        Verify that maybe_open_issue will create a new issue when a Resolved one
+        exists with the same name.
+        """
+        issue = G(Issue, status=IssueStatus.Resolved.value)
+        (issue2, created) = Issue.objects.maybe_open_issue(name=issue.name)
+        self.assertTrue(created)
+        self.assertEqual(IssueStatus.Open.value, Issue.objects.get(pk=issue2.pk).status)
+        self.assertEqual(2, Issue.objects.get(name=issue2.name))
+
+    def test_resolve_open_issue(self):
+        a = G(Assertion)
+        issue = G(Issue, name=a.name, status=IssueStatus.Open.value)
+        a._resolve_open_issue()
+        self.assertEqual(IssueStatus.Resolved.value, Issue.objects.get(pk=issue.pk).status)
 
 
 class ModelIssueManagerTests(TestCase):
@@ -108,9 +144,9 @@ class ModelIssueManagerTests(TestCase):
 
 
 class IssueTests(TestCase):
-    def test__unicode__(self):
+    def test__str__(self):
         i = Issue(name='an-issue', status=IssueStatus.Resolved.value)
-        self.assertEqual('Issue: an-issue - IssueStatus.Resolved', i.__unicode__())
+        self.assertEqual('Issue: an-issue - IssueStatus.Resolved', str(i))
 
     def test__is_open(self):
         i = N(Issue, status=IssueStatus.Open.value)
@@ -132,24 +168,36 @@ class IssueTests(TestCase):
 
 
 class IssueActionTests(TestCase):
-    def test__unicode__(self):
-        ia = N(IssueAction)
+    def test__str__(self):
+        issue = G(Issue, name='bar')
+        responder = G(Responder, watch_pattern='foo')
+        responder_action = G(ResponderAction, responder=responder)
+        issue_action = N(
+            IssueAction,
+            responder_action=responder_action,
+            action_issue_type=ContentType.objects.get_for_model(Issue),
+            action_issue_id=issue.id
+        )
         self.assertEqual(
-            u'IssueResponse: {self.issue.name} - {self.responder_action} - '
-            '{self.success} at {self.execution_time}'.format(self=ia),
-            ia.__unicode__())
+            'IssueResponse: {self.action_issue.name} - {self.responder_action} - '
+            '{self.success} at {self.execution_time}'.format(self=issue_action),
+            str(issue_action)
+        )
 
 
 class ResponderTests(TestCase):
-    def test__unicode__(self):
-        self.assertEqual('Responder: error-.*', Responder(watch_pattern='error-.*').__unicode__())
+    def test__str__(self):
+        self.assertEqual(
+            'Responder: error-.*',
+            str(Responder(watch_pattern='error-.*'))
+        )
 
     @patch('issue.models.load_function', spec_set=True)
     def test_respond(self, load_function):
         # Setup the scenario
         target_function = 'do'
         issue = G(Issue, name='error-42')
-        responder = G(Responder, issue=issue, watch_pattern='error-\d+')
+        responder = G(Responder, issue=issue, watch_pattern=r'error-\d+')
         G(ResponderAction, responder=responder, target_function=target_function, delay_sec=0)
 
         # Run the code
@@ -163,7 +211,7 @@ class ResponderTests(TestCase):
     def test_respond_ignores_non_watching_pattern(self, load_function):
         # Setup the scenario
         issue = G(Issue, name='success')
-        responder = G(Responder, issue=issue, watch_pattern='error-\d+')
+        responder = G(Responder, issue=issue, watch_pattern=r'error-\d+')
         G(ResponderAction, responder=responder, target_function='do')
 
         # Run the code
@@ -193,13 +241,18 @@ class ResponderTests(TestCase):
         # Setup the scenario
         now = datetime(2014, 8, 11, 15, 0, 0)
         delta = timedelta(minutes=30)
-        r = G(Responder)
-        ra = G(ResponderAction, responder=r, delay_sec=delta.total_seconds())
+        responder = G(Responder, allow_retry=False)
+        responder_action = G(ResponderAction, responder=responder, delay_sec=delta.total_seconds())
         issue = G(Issue, creation_time=now - (delta * 2))
-        G(IssueAction, issue=issue, responder_action=ra)
+        G(
+            IssueAction,
+            action_issue_id=issue.id,
+            action_issue_type=ContentType.objects.get_for_model(Issue),
+            responder_action=responder_action
+        )
 
         # Run the code and verify expectation
-        self.assertFalse(r._get_pending_actions_for_issue(issue).exists())
+        self.assertFalse(responder._get_pending_actions_for_issue(issue).exists())
 
     @patch('issue.models.load_function', spec_set=True)
     def test__execute_all_success(self, load_function):
@@ -236,9 +289,9 @@ class ResponderTests(TestCase):
         self.assertTrue(self.do_call_time < self.do_2_call_time)
         self.assertTrue(self.do_2_call_time < self.do_3_call_time)
 
-        self.assertTrue(IssueAction.objects.filter(issue=issue, responder_action=ra).exists())
-        self.assertTrue(IssueAction.objects.filter(issue=issue, responder_action=ra2).exists())
-        self.assertTrue(IssueAction.objects.filter(issue=issue, responder_action=ra3).exists())
+        self.assertTrue(IssueAction.objects.filter(action_issue_id=issue.id, responder_action=ra).exists())
+        self.assertTrue(IssueAction.objects.filter(action_issue_id=issue.id, responder_action=ra2).exists())
+        self.assertTrue(IssueAction.objects.filter(action_issue_id=issue.id, responder_action=ra3).exists())
 
     @patch('issue.models.load_function', spec_set=True)
     def test__execute_stops_when_some_actions_are_not_yet_executable(self, load_function):
@@ -275,9 +328,9 @@ class ResponderTests(TestCase):
         self.assertTrue(self.do_call_time < self.do_2_call_time)
         self.assertIsNone(self.do_3_call_time)
 
-        self.assertTrue(IssueAction.objects.filter(issue=issue, responder_action=ra).exists())
-        self.assertTrue(IssueAction.objects.filter(issue=issue, responder_action=ra2).exists())
-        self.assertTrue(IssueAction.objects.filter(issue=issue, responder_action=ra3).exists())
+        self.assertTrue(IssueAction.objects.filter(action_issue_id=issue.id, responder_action=ra).exists())
+        self.assertTrue(IssueAction.objects.filter(action_issue_id=issue.id, responder_action=ra2).exists())
+        self.assertTrue(IssueAction.objects.filter(action_issue_id=issue.id, responder_action=ra3).exists())
 
     @freeze_time(datetime(2014, 8, 13, 12))
     @patch('issue.models.load_function', spec_set=True)
@@ -286,9 +339,19 @@ class ResponderTests(TestCase):
         delta = timedelta(seconds=30)
         issue = G(Issue, creation_time=datetime.utcnow() - (2 * delta))
         responder = G(Responder, issue=issue)
-        ra = G(ResponderAction, responder=responder, delay_sec=0, target_function='do_1')
-        ra2 = G(ResponderAction, responder=responder, delay_sec=delta.total_seconds(), target_function='do_2')
-        G(IssueAction, issue=issue, responder_action=ra)
+        responder_action = G(ResponderAction, responder=responder, delay_sec=0, target_function='do_1')
+        responder_action2 = G(
+            ResponderAction,
+            responder=responder,
+            delay_sec=delta.total_seconds(),
+            target_function='do_2'
+        )
+        G(
+            IssueAction,
+            action_issue_id=issue.id,
+            action_issue_type=ContentType.objects.get_for_model(Issue),
+            responder_action=responder_action
+        )
 
         self.do_called = False
         self.do_2_called = False
@@ -310,8 +373,18 @@ class ResponderTests(TestCase):
         self.assertFalse(self.do_called)
         self.assertTrue(self.do_2_called)
 
-        self.assertTrue(IssueAction.objects.filter(issue=issue, responder_action=ra).exists())
-        self.assertTrue(IssueAction.objects.filter(issue=issue, responder_action=ra2).exists())
+        self.assertTrue(
+            IssueAction.objects.filter(
+                action_issue_id=issue.id,
+                responder_action=responder_action
+            ).exists()
+        )
+        self.assertTrue(
+            IssueAction.objects.filter(
+                action_issue_id=issue.id,
+                responder_action=responder_action2
+            ).exists()
+        )
 
     @patch('issue.models.load_function', spec_set=True)
     def test__execute_failure_does_not_stop_other_actions(self, load_function):
@@ -349,21 +422,74 @@ class ResponderTests(TestCase):
         self.assertTrue(self.do_call_time < self.do_2_call_time)
         self.assertTrue(self.do_2_call_time < self.do_3_call_time)
 
-        self.assertTrue(IssueAction.objects.filter(issue=issue, responder_action=ra).exists())
-        self.assertTrue(IssueAction.objects.filter(issue=issue, responder_action=ra2).exists())
-        self.assertTrue(IssueAction.objects.filter(issue=issue, responder_action=ra3).exists())
+        self.assertTrue(IssueAction.objects.filter(action_issue_id=issue.id, responder_action=ra).exists())
+        self.assertTrue(IssueAction.objects.filter(action_issue_id=issue.id, responder_action=ra2).exists())
+        self.assertTrue(IssueAction.objects.filter(action_issue_id=issue.id, responder_action=ra3).exists())
         self.assertEqual(
             json.dumps(str(Exception('what-an-exceptional-message'))),
-            IssueAction.objects.get(issue=issue, responder_action=ra2).details)
+            IssueAction.objects.get(action_issue_id=issue.id, responder_action=ra2).details)
+
+    @patch('issue.models.load_function', spec_set=True)
+    def test_allow_retry(self, load_function):
+        """
+        Confirms that responders that are configured to allow retry will not be restricted
+        by previous IssueActions
+        """
+        # Setup the scenario
+        issue = G(Issue)
+        responder = G(Responder, allow_retry=True)
+        # Note: we don't care what the target_function path is since we patch the load_function function
+        responder_action = G(ResponderAction, responder=responder, delay_sec=0)
+        responder_action2 = G(ResponderAction, responder=responder, delay_sec=0)
+
+        self.do_call_time = None
+        self.do_2_call_time = None
+
+        def do_1(*args, **kwargs):
+            self.do_call_time = datetime.utcnow()
+            return True
+
+        def do_2(*args, **kwargs):
+            self.do_2_call_time = datetime.utcnow()
+            return True
+
+        load_function.side_effect = [do_1, do_2]
+
+        # attempt to respond
+        responder._execute(issue)
+
+        # Verify expectations
+        self.assertEqual(2, IssueAction.objects.count())
+        self.assertTrue(
+            IssueAction.objects.filter(
+                action_issue_id=issue.id,
+                responder_action=responder_action
+            ).exists()
+        )
+        self.assertTrue(
+            IssueAction.objects.filter(
+                action_issue_id=issue.id,
+                responder_action=responder_action2
+            ).exists()
+        )
+
+        # run again
+        responder._execute(issue)
+        self.assertEqual(4, IssueAction.objects.count())
 
 
 class ResponderActionTests(TestCase):
-    def test__unicode(self):
-        r = G(ResponderAction)
+    def test__str__(self):
+        responder = G(Responder, watch_pattern='foo')
+        responder_action = G(ResponderAction, responder=responder)
         self.assertEqual(
             'ResponderAction: {responder} - {target_function} - {function_kwargs}'.format(
-                responder=r.responder, target_function=r.target_function, function_kwargs=r.function_kwargs),
-            r.__unicode__())
+                responder=responder_action.responder,
+                target_function=responder_action.target_function,
+                function_kwargs=responder_action.function_kwargs
+            ),
+            str(responder_action)
+        )
 
     def test_is_time_to_execute(self):
         # Setup the scenario
@@ -413,7 +539,7 @@ class ResponderActionTests(TestCase):
         load_function.assert_called_with(target_function)
         load_function.return_value.assert_called_with(issue, foo='bar')
 
-        self.assertTrue(IssueAction.objects.filter(issue=issue, **expected_issue_action_kwargs).exists())
+        self.assertTrue(IssueAction.objects.filter(action_issue_id=issue.id, **expected_issue_action_kwargs).exists())
         # The 'None' that is stored as the details is first json encoded
         self.assertEqual(json.dumps(None), IssueAction.objects.get().details)
 
@@ -443,7 +569,7 @@ class ResponderActionTests(TestCase):
         load_function.assert_called_with(target_function)
         load_function.return_value.assert_called_with(issue, foo='bar')
 
-        self.assertTrue(IssueAction.objects.filter(issue=issue, **expected_issue_action_kwargs).exists())
+        self.assertTrue(IssueAction.objects.filter(action_issue_id=issue.id, **expected_issue_action_kwargs).exists())
         self.assertEqual(json.dumps(str(Exception('what-an-exceptional-message'))), IssueAction.objects.get().details)
 
 
@@ -458,10 +584,10 @@ class AssertionTests(TestCase):
 
         assertion = G(Assertion, check_function='issue.tests.model_tests.load_function')
 
-        self.assertTrue(assertion.check())
+        self.assertTrue(assertion.check_assertion())
         self.assertTrue(resolve_open_issue.called)
 
-    @patch.object(Assertion, '_open_or_update_issue', spec_set=True)
+    @patch.object(Assertion, '_open_or_update_issue')
     @patch('issue.models.load_function', spec_set=True)
     def test_check_when_all_is_not_well(self, load_function, open_or_update_issue):
         issue_details = {
@@ -471,7 +597,7 @@ class AssertionTests(TestCase):
 
         assertion = G(Assertion, check_function='issue.tests.model_tests.load_function')
 
-        self.assertFalse(assertion.check())
+        self.assertFalse(assertion.check_assertion())
         open_or_update_issue.assert_called_with(details=issue_details)
 
     def test__open_or_update_issue_when_none_exists(self):
@@ -487,9 +613,9 @@ class AssertionTests(TestCase):
 
     def test__open_or_update_issue_when_it_is_marked_as_resolved(self):
         a = G(Assertion)
-        issue = G(Issue, name=a.name, status=IssueStatus.Resolved.value)
-        a._open_or_update_issue({})
-        self.assertEqual(IssueStatus.Open.value, Issue.objects.get(pk=issue.pk).status)
+        G(Issue, name=a.name, status=IssueStatus.Resolved.value)
+        issue2 = a._open_or_update_issue({})
+        self.assertEqual(IssueStatus.Open.value, Issue.objects.get(pk=issue2.pk).status)
 
     def test_resolve_open_issue(self):
         a = G(Assertion)
@@ -516,7 +642,7 @@ class ModelAssertionTests(TestCase):
             check_function='issue.tests.model_tests.is_even_number')
 
         # Run the code
-        r = ma.check()
+        r = ma.check_assertion()
 
         # Verify expectations
         self.assertTrue(r)
@@ -532,7 +658,7 @@ class ModelAssertionTests(TestCase):
             check_function='issue.tests.model_tests.is_even_number')
 
         # Run the code
-        r = ma.check()
+        r = ma.check_assertion()
 
         # Verify expectations
         self.assertFalse(r)
